from tensorflow.keras.optimizers import *
from tensorflow.keras.layers import *
from tensorflow.keras.models import *
import tensorflow as tf
import numpy as np
import importlib


class KerasCNNGrid:

    def __init__(self, settings, hyperparameters, fixed_model=False, loss_function=None):
        self.settings = settings
        self.hyperparameters = hyperparameters
        self.model_structure = settings["grid_search"]["structure"]
        self.fixed_model = fixed_model
        self.loss_function = loss_function

    def cnn_grid_search(self):

        conv_layers = self.hyperparameters["conv_layers"] if "conv_layers" in self.hyperparameters else 1
        kernels = []
        strides = []
        filters = []
        pooling_types = []
        pooling_sizes = []
        pooling_strides = []
        for conv_layer in range(1, conv_layers + 1):
            kernels = self.get_convolution_hyperparameter(conv_layer, kernels, "kernel")
            strides = self.get_convolution_hyperparameter(conv_layer, strides, "stride")
            filters = self.get_convolution_hyperparameter(conv_layer, filters, "filters")
            pooling_sizes = self.get_pooling_hyperparameter(conv_layer, pooling_sizes, "size")
            pooling_strides = self.get_pooling_hyperparameter(conv_layer, pooling_strides, "stride")
            pooling_types = self.get_pooling_type(conv_layer, pooling_types)
        neurons = self.hyperparameters["neurons"] if "neurons" in self.hyperparameters else 50
        layers = self.hyperparameters["layers"] if "layers" in self.hyperparameters else 5
        activation = self.hyperparameters["activation"] if "activation" in self.hyperparameters else "relu"
        learning_rate = self.hyperparameters["learning_rate"] if "learning_rate" in self.hyperparameters else 0.001
        optimizer = self.hyperparameters["optimizer"] if "optimizer" in self.hyperparameters else "Adam"
        dropout_rate = self.hyperparameters["dropout_rate"] if "dropout_rate" in self.hyperparameters else 0
        random_seed = self.hyperparameters["seed"] if self.fixed_model else np.random.randint(1048576)

        hp = {'conv_layers': conv_layers}
        for conv_layer in range(1, conv_layers + 1):
            hp["kernel_{}".format(conv_layer)] = kernels[conv_layer - 1]
            hp["stride_{}".format(conv_layer)] = strides[conv_layer - 1]
            hp["filters_{}".format(conv_layer)] = filters[conv_layer - 1]
            hp["pooling_size_{}".format(conv_layer)] = pooling_sizes[conv_layer - 1]
            hp["pooling_stride_{}".format(conv_layer)] = pooling_strides[conv_layer - 1]
            hp["pooling_type_{}".format(conv_layer)] = pooling_types[conv_layer - 1]
        hp["neurons"] = neurons
        hp["layers"] = layers
        hp["activation"] = activation
        hp["learning_rate"] = learning_rate
        if self.model_structure["use_dropout_after_dense_layer"] or self.model_structure["use_dropout_before_dense_layer"]:
            hp['dropout_rate'] = dropout_rate
        hp["optimizer"] = optimizer
        hp["seed"] = random_seed

        tf.random.set_seed(random_seed)

        model = Sequential()
        for conv_layer in range(1, conv_layers + 1):
            if conv_layer == 1:

                if self.model_structure["use_pooling_before_first_convolution"]:

                    if "pooling_size_0" in self.hyperparameters and "pooling_stride_0" in self.hyperparameters and "pooling_type_0" in self.hyperparameters:

                        pooling_sizes_0 = self.hyperparameters["pooling_size_0"]
                        pooling_stride_0 = self.hyperparameters["pooling_stride_0"]
                        pooling_type_0 = self.hyperparameters["pooling_type_0"]

                        if pooling_type_0 == "average":
                            model.add(AveragePooling1D(pool_size=pooling_sizes_0, strides=pooling_stride_0),
                                      input_shape=(self.settings["number_of_samples"], 1))
                        else:
                            model.add(MaxPool1D(pool_size=pooling_sizes_0, strides=pooling_stride_0,
                                                input_shape=(self.settings["number_of_samples"], 1)))

                        hp["pooling_size_0"] = pooling_sizes_0
                        hp["pooling_stride_0"] = pooling_stride_0
                        hp["pooling_type_0"] = pooling_type_0

                    model.add(
                        Conv1D(kernel_size=kernels[conv_layer - 1], strides=strides[conv_layer - 1], filters=filters[conv_layer - 1],
                               activation=activation, padding="same"))
                else:
                    model.add(
                        Conv1D(kernel_size=kernels[conv_layer - 1], strides=strides[conv_layer - 1], filters=filters[conv_layer - 1],
                               activation=activation, padding="same", input_shape=(self.settings["number_of_samples"], 1)))
            else:
                model.add(
                    Conv1D(kernel_size=kernels[conv_layer - 1], strides=strides[conv_layer - 1], filters=filters[conv_layer - 1],
                           activation=activation, padding="same"))
            if self.model_structure["use_batch_norm_after_convolution"] or (
                    self.model_structure["use_batch_norm_before_pooling"] and self.model_structure["use_pooling_after_convolution"]):
                model.add(BatchNormalization())
            if self.model_structure["use_pooling_after_convolution"]:
                if pooling_types[conv_layer - 1] == "average":
                    model.add(AveragePooling1D(pool_size=pooling_sizes[conv_layer - 1], strides=pooling_strides[conv_layer - 1]))
                else:
                    model.add(MaxPool1D(pool_size=pooling_sizes[conv_layer - 1], strides=pooling_strides[conv_layer - 1]))
            if self.model_structure["use_batch_norm_after_pooling"]:
                model.add(BatchNormalization())
        model.add(Flatten())
        for layer_index in range(layers):
            if self.model_structure["use_dropout_before_dense_layer"]:
                model.add(Dropout(dropout_rate))
            model.add(Dense(neurons, activation=activation))
            if self.model_structure["use_dropout_after_dense_layer"]:
                model.add(Dropout(dropout_rate))
        if self.model_structure["use_dropout_before_dense_layer"]:
            model.add(Dropout(dropout_rate))
        model.add(Dense(self.settings["classes"], activation='softmax'))
        model.summary()

        if self.loss_function is not None:
            class_inverted = self.loss_function["class"][::-1]
            module_name = class_inverted.partition(".")[2][::-1]
            loss_function_name = class_inverted.partition(".")[0][::-1]
            module_name = importlib.import_module(module_name)
            custom_loss_class = getattr(module_name, loss_function_name)
            loss = custom_loss_class(self.settings, parameters=self.loss_function['parameters'])
            model.compile(loss=loss, optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        else:
            model.compile(loss='categorical_crossentropy', optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        return model, hp

    def get_optimizer(self, optimizer, learning_rate):
        if optimizer == "Adam":
            return Adam(lr=learning_rate)
        elif optimizer == "RMSprop":
            return RMSprop(lr=learning_rate)
        elif optimizer == "Adadelta":
            return Adadelta(lr=learning_rate)
        elif optimizer == "Adagrad":
            return Adagrad(lr=learning_rate)
        elif optimizer == "SGD":
            return SGD(lr=learning_rate, momentum=0.9, nesterov=True)
        else:
            return Adam(lr=learning_rate)

    def get_convolution_hyperparameter(self, conv_layer, hp_list, hp_type):

        label = f"{hp_type}_{conv_layer}"

        if conv_layer > 1:
            if self.hyperparameters[label] == "equal_from_previous_convolution":
                new_hp = hp_list[conv_layer - 2]
            elif self.hyperparameters[label] == "double_from_previous_convolution":
                new_hp = hp_list[conv_layer - 2] * 2
            elif self.hyperparameters[label] == "half_from_previous_convolution":
                new_hp = int(hp_list[conv_layer - 2] / 2) if hp_list[conv_layer - 2] > 1 else 1
            else:
                new_hp = self.hyperparameters[label]
        else:
            new_hp = self.hyperparameters[label]
        hp_list.append(new_hp)
        return hp_list

    def get_pooling_hyperparameter(self, conv_layer, hp_list, hp_type):

        label = f"pooling_{hp_type}_{conv_layer}"

        if conv_layer > 1:
            if self.hyperparameters[label] == "equal_from_previous_pooling":
                value = hp_list[conv_layer - 2]
            elif self.hyperparameters[label] == "double_from_previous_pooling":
                value = hp_list[conv_layer - 2] * 2
            elif self.hyperparameters[label] == "half_from_previous_pooling":
                value = int(hp_list[conv_layer - 2] / 2) if hp_list[conv_layer - 2] > 1 else 1
            else:
                value = self.hyperparameters[label]
        else:
            value = self.hyperparameters[label]
        hp_list.append(value)
        return hp_list

    def get_pooling_type(self, conv_layer, hp_list):
        if conv_layer > 1:
            if self.hyperparameters[f"pooling_type_{conv_layer}"] == "equal_from_previous_pooling":
                hp_list.append(hp_list[conv_layer - 1])
            else:
                hp_list.append(self.hyperparameters[f"pooling_type_{conv_layer}"])
        else:
            hp_list.append(self.hyperparameters[f"pooling_type_{conv_layer}"])
        return hp_list



from aisy_sca.analysis.SingleModel import SingleModel
from aisy_sca.analysis.MultiModel import MultiModel
from aisy_sca.AnalysisDatabaseControls import *
from aisy_sca.LeakageModelControls import *


class SingleProcess:

    def __init__(self, dataset, settings, models, da_function, labels_key_guesses_attack_set=None, labels_key_guesses_validation_set=None,
                 random_states=None):
        self.dataset = dataset
        self.settings = settings
        self.models = models
        self.da_function = da_function
        self.custom_callbacks = None
        self.labels_key_guesses_attack_set = labels_key_guesses_attack_set
        self.labels_key_guesses_validation_set = labels_key_guesses_validation_set
        self.random_states = random_states

    def run(self):

        """ Create analysis in database """
        analysis_db_controls = AnalysisDatabaseControls(self.settings)
        self.settings["analysis_id"] = analysis_db_controls.analysis_id
        analysis_db_controls.save_leakage_model_to_database()

        """ Generate list of labels for all key guesses for validation and attack sets """
        if self.labels_key_guesses_attack_set is None and self.labels_key_guesses_validation_set is None:
            lm_control = LeakageModelControls(self.settings)
            self.labels_key_guesses_attack_set, self.labels_key_guesses_validation_set = lm_control.compute_labels_key_guesses(self.dataset)

        """ If number of neural networks is larger than 1, starts multi-model process. Otherwise, runs single-model process """
        if len(self.models) > 1:
            multi_model = MultiModel(self.settings, self.dataset, self.da_function, random_states=self.random_states)
            multi_model.labels_key_guesses_attack_set = self.labels_key_guesses_attack_set
            multi_model.labels_key_guesses_validation_set = self.labels_key_guesses_validation_set
            multi_model.analysis_db_controls = analysis_db_controls
            multi_model.run(self.models)
            self.custom_callbacks = multi_model.custom_callbacks
        else:
            single_model = SingleModel(self.settings, self.dataset, self.da_function, random_states=self.random_states)
            single_model.labels_key_guesses_attack_set = self.labels_key_guesses_attack_set
            single_model.labels_key_guesses_validation_set = self.labels_key_guesses_validation_set
            single_model.analysis_db_controls = analysis_db_controls
            single_model.run(self.models["0"])
            self.custom_callbacks = single_model.custom_callbacks



class Dataset:

    def __init__(self, x_profiling, y_profiling, x_attack, y_attack, x_validation=None, y_validation=None, plaintext_profiling=None,
                 plaintext_attack=None, ciphertext_profiling=None, ciphertext_attack=None, key_profiling=None, key_attack=None,
                 plaintext_validation=None, ciphertext_validation=None, key_validation=None):
        """ set profiling, validation and attack traces"""
        self.x_profiling = x_profiling
        self.x_attack = x_attack
        self.x_validation = x_validation

        """ set profiling, validation and attack labels"""
        self.y_profiling = y_profiling
        self.y_attack = y_attack
        self.y_validation = y_validation

        """ set profiling, validation and attack plaintext"""
        self.plaintext_profiling = plaintext_profiling
        self.plaintext_attack = plaintext_attack
        self.plaintext_validation = plaintext_validation

        """ set profiling, validation and attack ciphertext"""
        self.ciphertext_profiling = ciphertext_profiling
        self.ciphertext_attack = ciphertext_attack
        self.ciphertext_validation = ciphertext_validation

        """ set profiling, validation and attack keys"""
        self.key_profiling = key_profiling
        self.key_attack = key_attack
        self.key_validation = key_validation

    def reshape_for_cnn(self):
        """ reshape traces for CNNs """
        self.x_profiling = self.x_profiling.reshape((self.x_profiling.shape[0], self.x_profiling.shape[1], 1))
        if self.x_validation is not None:
            self.x_validation = self.x_validation.reshape((self.x_validation.shape[0], self.x_validation.shape[1], 1))
        self.x_attack = self.x_attack.reshape((self.x_attack.shape[0], self.x_attack.shape[1], 1))

    def reshape_for_mlp(self):
        """ reshape traces for CNNs """
        self.x_profiling = self.x_profiling.reshape((self.x_profiling.shape[0], self.x_profiling.shape[1]))
        if self.x_validation is not None:
            self.x_validation = self.x_validation.reshape((self.x_validation.shape[0], self.x_validation.shape[1]))
        self.x_attack = self.x_attack.reshape((self.x_attack.shape[0], self.x_attack.shape[1]))



import setuptools

setuptools.setup(
    name="aisy_sca",
    version="0.2.7",
    author="Guilherme Perin",
    author_email="guilhermeperin7@gmail.com",
    description="Deep Learning-based Framework for Side Channel Analysis",
    long_description="""
    # AISY - Deep Learning-based Framework for Side Channel Analysis

    Welcome to the first deep learning-based side-channel analysis framework.
    This AISY Framework was developed by AISYLab from TU Delft.

    If you use our framework, please consider citing:

    @misc{AISY_Framework,
        author = {Guilherme Perin and Lichao Wu and Stjepan Picek},
        title  = {{AISY - Deep Learning-based Framework for Side-Channel Analysis}},
        howpublished = {AISyLab repository},
        note   = {{\\url{https://github.com/AISyLab/AISY_Framework}}},
        year   = {2021}
    }
    """,
    long_description_content_type='text/markdown',
    keywords="side-channel analysis deep learning profiled attacks",
    packages=['aisy_sca', 'aisy_sca/analysis', 'aisy_sca/crypto', 'aisy_sca/crypto/aes', 'aisy_sca/datasets', 'aisy_sca/callbacks',
              'aisy_sca/metrics', 'aisy_sca/optimization', 'aisy_sca/sca', 'aisy_sca/utils', 'tests'],
    url="https://aisylab.github.io/AISY_docs/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"]
)



from aisy_sca.utils import Utils
from aisy_sca.datasets import DatasetControls
from aisy_sca.AnalysisControls import *
from aisy_sca.analysis.SingleProcess import SingleProcess
from aisy_sca.analysis.HyperparameterSearchProcess import HyperparameterSearchProcess
from aisy_sca.LeakageModelControls import *
from numpy.random import seed as numpy_seed
import tensorflow.keras.backend as backend
import tensorflow as tf
import inspect
import importlib
import time


class Aisy:

    def __init__(self):

        """ variables for analysis"""
        self.settings = {}
        self.dataset = None
        self.models = {}
        self.custom_callbacks = None
        self.loss_function = None

        """ variables from external datasets """
        self.labels_key_guesses_attack_set = None
        self.labels_key_guesses_validation_set = None

        """ reproducibility """
        self.random_states = None
        self.hyperparameters_reproducible = None

    def run(self, key_rank_executions=100, key_rank_report_interval=10, key_rank_attack_traces=1000, visualization=None,
            data_augmentation=None, ensemble=None, grid_search=None, random_search=None, early_stopping=None, confusion_matrix=False,
            callbacks=None, profiling_analyzer=None):

        """
        Main AISY framework function. This function runs neural network training for profiled side-channel analysis in a known-key setting.
        """

        self.settings["key_rank_executions"] = key_rank_executions
        self.settings["key_rank_attack_traces"] = key_rank_attack_traces
        self.settings["key_rank_report_interval"] = key_rank_report_interval
        self.settings["timestamp"] = str(time.time()).replace(".", "")

        analysis_controls = AnalysisControls(self.settings, self.models)
        analysis_controls.update_settings(random_search, grid_search, visualization, early_stopping, data_augmentation, callbacks, ensemble,
                                          confusion_matrix, profiling_analyzer)
        analysis_controls.check_errors()
        da_function = data_augmentation[0] if data_augmentation is not None else None

        """ Load data sets """
        if self.dataset is None:
            ds_controls = DatasetControls(self.settings)
            ds_controls.read()
            self.dataset = ds_controls.get_dataset()

        """ Run main process """
        if self.settings["hyperparameter_search"]:
            hyperparameter_search_process = HyperparameterSearchProcess(
                self.dataset, self.settings, self.models, da_function, self.hyperparameters_reproducible,
                labels_key_guesses_attack_set=self.labels_key_guesses_attack_set,
                labels_key_guesses_validation_set=self.labels_key_guesses_validation_set,
                loss_function=self.loss_function, random_states=self.random_states)
            hyperparameter_search_process.run()

            """ Get custom callbacks """
            self.custom_callbacks = hyperparameter_search_process.custom_callbacks
        else:
            single_process = SingleProcess(
                self.dataset, self.settings, self.models, da_function,
                labels_key_guesses_attack_set=self.labels_key_guesses_attack_set,
                labels_key_guesses_validation_set=self.labels_key_guesses_validation_set,
                random_states=self.random_states)
            single_process.run()

            """ Get custom callbacks """
            self.custom_callbacks = single_process.custom_callbacks

    def set_datasets_root_folder(self, datasets_root_folder):
        self.settings["datasets_root_folder"] = Utils().adapt_folder_path(datasets_root_folder)

    def set_database_root_folder(self, database_root_folder):
        self.settings["database_root_folder"] = Utils().adapt_folder_path(database_root_folder)

    def set_resources_root_folder(self, resources_root_folder):
        self.settings["resources_root_folder"] = Utils().adapt_folder_path(resources_root_folder)

    def set_dataset_filename(self, dataset_filename):
        self.settings["filename"] = dataset_filename

    def set_first_sample(self, first_sample):
        self.settings["first_sample"] = first_sample

    def set_number_of_samples(self, number_of_samples):
        self.settings["number_of_samples"] = number_of_samples

    def set_number_of_profiling_traces(self, profiling_traces):
        self.settings["number_of_profiling_traces"] = profiling_traces

    def set_number_of_attack_traces(self, attack_traces):
        self.settings["number_of_attack_traces"] = attack_traces

    def set_key(self, key):
        self.settings["key"] = key

    def set_dataset(self, target_dict, dataset=None):
        for key, value in target_dict.items():
            self.settings[str(key)] = value
        if dataset is not None:
            self.dataset = dataset

    def set_database_name(self, database_name):
        self.settings["database_name"] = database_name

    def set_batch_size(self, batch_size):
        self.settings["batch_size"] = batch_size

    def set_epochs(self, epochs):
        self.settings["epochs"] = epochs

    def set_hyperparameters_reproducible(self, hyperparameters_reproducible):
        self.hyperparameters_reproducible = hyperparameters_reproducible

    def set_random_states(self, random_states):
        self.random_states = random_states

    def get_models(self):
        return self.models

    def set_model_weights(self, weights, model_index=0):
        self.models[f"{model_index}"].set_weights(weights)

    def set_neural_network(self, model, name=None):
        self.add_neural_network(model, name=name)

    def add_neural_network(self, model, name=None, seed=None):
        if not bool(self.models):
            self.settings["models"] = {}
        model_index = len(self.models)
        if inspect.isfunction(model):
            model_name = model.__name__ if name is None else name
            tf_random_seed = np.random.randint(1048576) if seed is None else seed
            tf.random.set_seed(tf_random_seed)
            method_name = model.__name__
            model = model(self.settings["classes"], self.settings["number_of_samples"])
            self.__set_model_attributes(model_index, model_name, method_name, tf_random_seed, model)
        else:
            self.__set_model_attributes(model_index, model.name if name is None else name, model.name, seed, model)

    def __set_model_attributes(self, model_index, model_name, method_name, seed, model):

        self.models[f"{model_index}"] = {}
        self.models[f"{model_index}"]["model_name"] = model_name
        self.models[f"{model_index}"]["method_name"] = method_name
        self.models[f"{model_index}"]["seed"] = seed
        self.models[f"{model_index}"]["model"] = model
        self.models[f"{model_index}"]["index"] = model_index

        self.settings["models"][f"{model_index}"] = {}
        self.settings["models"][f"{model_index}"]["model_name"] = model_name
        self.settings["models"][f"{model_index}"]["method_name"] = method_name
        self.settings["models"][f"{model_index}"]["seed"] = seed
        self.settings["models"][f"{model_index}"]["index"] = model_index

    def set_loss_function(self, loss_dict, model_args=None):

        self.loss_function = loss_dict

        class_inverted = loss_dict["class"][::-1]
        module_name = class_inverted.partition(".")[2][::-1]
        loss_function_name = class_inverted.partition(".")[0][::-1]
        module_name = importlib.import_module(module_name)
        custom_loss_class = getattr(module_name, loss_function_name)

        if model_args is not None:
            if len(model_args) != len(self.models):
                print("Expected args to be a list with same amount of models.")
                return

        for model_index, model in self.models.items():
            learning_rate = backend.eval(model["model"].optimizer.lr)
            optimizer = model["model"].optimizer.__class__
            if model_args is not None:
                m_args = model_args[int(model_index)]
            else:
                m_args = None
            model["model"].compile(
                loss=custom_loss_class(self.settings, model_args=m_args, parameters=loss_dict['parameters']),
                optimizer=optimizer(learning_rate=learning_rate), metrics=['accuracy'])

    def set_aes_leakage_model(self, leakage_model="HW", bit=0, byte=0, round=1, round_first=1, round_second=1, cipher="AES128",
                              target_state="Sbox", target_state_first="Sbox", target_state_second="Sbox",
                              direction="Encryption", attack_direction="input"):
        lm_control = LeakageModelControls(self.settings)
        self.settings = lm_control.set(leakage_model, bit, byte, round, round_first, round_second, cipher, target_state, target_state_first,
                                       target_state_second, direction, attack_direction)
        self.settings["good_key"] = bytearray.fromhex(self.settings["key"])[self.settings["leakage_model"]["byte"]]

    def set_good_key(self, key):
        self.settings["good_key"] = int(key)

    def set_classes(self, classes):
        self.settings["classes"] = classes

    def set_labels_key_guesses_attack_set(self, labels_key_guesses_attack_set):
        self.labels_key_guesses_attack_set = labels_key_guesses_attack_set

    def set_labels_key_guesses_validation_set(self, labels_key_guesses_validation_set):
        self.labels_key_guesses_validation_set = labels_key_guesses_validation_set

    def get_custom_callbacks(self):
        return self.custom_callbacks



from aisy_sca.crypto.aes.intermediates_decrypt import *
from aisy_sca.crypto.aes.intermediates_encrypt import *


def repeat_key_candidate(keys, nb_traces):
    """ repeat key if argument keys is a single key candidate (for GE and SR computations)"""
    if np.array(keys).ndim == 1:
        return np.full([nb_traces, len(keys)], keys)
    else:
        return keys


def get_round_keys(keys):
    keys = np.array(keys, dtype='uint8')
    if np.all(keys == keys[0]):
        """ If all keys are equal, then compute round keys for one key only """
        round_keys = expand_key(list(keys[0]))
        return np.full([len(keys), len(round_keys), len(round_keys[0])], round_keys)
    else:
        return [expand_key(list(key)) for key in keys]


def aes_intermediates(plaintexts, ciphertexts, keys, leakage_model):
    if "input" in leakage_model["attack_direction"]:
        if "Encryption" in leakage_model["direction"]:
            keys = repeat_key_candidate(keys, len(plaintexts))
            intermediate_values = aes_intermediates_encrypt_from_input(plaintexts, keys, leakage_model)
        else:
            keys = repeat_key_candidate(keys, len(ciphertexts))
            intermediate_values = aes_intermediates_decrypt_from_input(ciphertexts, get_round_keys(keys), leakage_model)
    else:
        if "Encryption" in leakage_model["direction"]:
            keys = repeat_key_candidate(keys, len(ciphertexts))
            intermediate_values = aes_intermediates_encrypt_from_output(ciphertexts, get_round_keys(keys), leakage_model)
        else:
            keys = repeat_key_candidate(keys, len(plaintexts))
            intermediate_values = aes_intermediates_decrypt_from_output(plaintexts, keys, leakage_model)

    if leakage_model["leakage_model"] in ["HW", "HD"]:
        return [bin(iv).count("1") for iv in intermediate_values]
    elif leakage_model["leakage_model"] == "bit":
        return [int(bin(iv >> leakage_model["bit"])[len(bin(iv >> leakage_model["bit"])) - 1]) for iv in intermediate_values]
    else:
        return intermediate_values



from tensorflow.keras.optimizers import *
from tensorflow.keras.layers import *
from tensorflow.keras.models import *
import tensorflow as tf
import numpy as np
import importlib


class KerasMLPGrid:

    def __init__(self, settings, hyperparameters, fixed_model=False, loss_function=None):
        self.settings = settings
        self.hyperparameters = hyperparameters
        self.model_structure = settings["grid_search"]["structure"]
        self.fixed_model = fixed_model
        self.loss_function = loss_function

    def mlp_grid_search(self):

        neurons = self.hyperparameters["neurons"] if "neurons" in self.hyperparameters else 50
        layers = self.hyperparameters["layers"] if "layers" in self.hyperparameters else 5
        activation = self.hyperparameters["activation"] if "activation" in self.hyperparameters else "relu"
        learning_rate = self.hyperparameters["learning_rate"] if "learning_rate" in self.hyperparameters else 0.001
        optimizer = self.hyperparameters["optimizer"] if "optimizer" in self.hyperparameters else "Adam"
        dropout_rate = self.hyperparameters["dropout_rate"] if "dropout_rate" in self.hyperparameters else 0
        random_seed = self.hyperparameters["seed"] if self.fixed_model else np.random.randint(1048576)

        hp = {
            'neurons': neurons,
            'layers': layers,
            'activation': activation,
            'learning_rate': learning_rate,
            'optimizer': optimizer,
            'seed': random_seed
        }

        if self.model_structure["use_dropout_after_dense_layer"] or self.model_structure["use_dropout_before_dense_layer"]:
            hp['dropout_rate'] = dropout_rate

        tf.random.set_seed(random_seed)

        model = Sequential()
        if self.model_structure["use_pooling_before_first_dense"]:
            if "pooling_size_0" in self.hyperparameters and "pooling_stride_0" in self.hyperparameters and "pooling_type_0" in self.hyperparameters:
                pooling_sizes_0 = self.hyperparameters["pooling_size_0"]
                pooling_stride_0 = self.hyperparameters["pooling_stride_0"]
                pooling_type_0 = self.hyperparameters["pooling_type_0"]
                if pooling_type_0 == "average":
                    model.add(AveragePooling1D(pool_size=pooling_sizes_0, strides=pooling_stride_0),
                              input_shape=(self.settings["number_of_samples"], 1))
                else:
                    model.add(MaxPool1D(pool_size=pooling_sizes_0, strides=pooling_stride_0,
                                        input_shape=(self.settings["number_of_samples"], 1)))
                model.add(Flatten())
                if self.model_structure["use_dropout_before_dense_layer"]:
                    model.add(Dropout(dropout_rate))
                model.add(Dense(neurons, activation=activation, input_shape=(self.settings["number_of_samples"],)))

                hp["pooling_size_0"] = pooling_sizes_0
                hp["pooling_stride_0"] = pooling_stride_0
                hp["pooling_type_0"] = pooling_type_0
        else:
            if self.model_structure["use_dropout_before_dense_layer"]:
                model.add(Dropout(dropout_rate))
            model.add(Dense(neurons, activation=activation, input_shape=(self.settings["number_of_samples"],)))
        if self.model_structure["use_dropout_after_dense_layer"]:
            model.add(Dropout(dropout_rate))
        for layer_index in range(layers - 1):
            if self.model_structure["use_dropout_before_dense_layer"]:
                model.add(Dropout(dropout_rate))
            model.add(Dense(neurons, activation=activation))
            if self.model_structure["use_dropout_after_dense_layer"]:
                model.add(Dropout(dropout_rate))
        if self.model_structure["use_dropout_before_dense_layer"]:
            model.add(Dropout(dropout_rate))
        model.add(Dense(self.settings["classes"], activation='softmax'))
        model.summary()

        if self.loss_function is not None:
            class_inverted = self.loss_function["class"][::-1]
            module_name = class_inverted.partition(".")[2][::-1]
            loss_function_name = class_inverted.partition(".")[0][::-1]
            module_name = importlib.import_module(module_name)
            custom_loss_class = getattr(module_name, loss_function_name)
            loss = custom_loss_class(self.settings, parameters=self.loss_function['parameters'])
            model.compile(loss=loss, optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        else:
            model.compile(loss='categorical_crossentropy', optimizer=self.get_optimizer(optimizer, learning_rate), metrics=['accuracy'])
        return model, hp

    def get_optimizer(self, optimizer, learning_rate):
        if optimizer == "Adam":
            return Adam(lr=learning_rate)
        elif optimizer == "RMSprop":
            return RMSprop(lr=learning_rate)
        elif optimizer == "Adadelta":
            return Adadelta(lr=learning_rate)
        elif optimizer == "Adagrad":
            return Adagrad(lr=learning_rate)
        elif optimizer == "SGD":
            return SGD(lr=learning_rate, momentum=0.9, nesterov=True)
        else:
            return Adam(lr=learning_rate)
